/*
 *  Copyright 2013-2050 Emmanuel BRUN (contact@amapj.fr)
 * 
 *  This file is part of AmapJ.
 *  
 *  AmapJ is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.

 *  AmapJ is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with AmapJ.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 */
 package fr.amapj.view.engine.popup.formpopup.fieldlink;

import java.util.ArrayList;
import java.util.List;

import org.vaadin.openesignforms.ckeditor.CKEditorTextField;

import com.vaadin.ui.AbstractField;
import com.vaadin.ui.ComboBox;
import com.vaadin.ui.FormLayout;
import com.vaadin.ui.TextField;

import fr.amapj.common.AmapjRuntimeException;
import fr.amapj.view.engine.popup.formpopup.ValidatorManager;
import fr.amapj.view.engine.popup.formpopup.validator.IValidator;
import fr.amapj.view.engine.popup.formpopup.validator.NotNullValidatorConditionnalFieldLink;

/**
 * Permet de créer un lien entre une combox box et l'activation / desactivation
 * d'un certain nombre d'élements liés
 */
public class FieldLink
{
	// Combo maitre 
	private ComboBox box;
	
	// Liste des éléments liés
	private List<FieldAndDefaultValue<?>> fields = new ArrayList<>(); 
	
	// Validateur qui pourra être appliqué sur les searchers si besoin 
	private NotNullValidatorConditionnalFieldLink notNull;
	
	// Liste des valeurs activant les elements
	private List<Enum<?>> actives = new ArrayList<>();

	private ValidatorManager validatorManager;
	
	// Etat courant 
	private boolean currentState;
	
	private FieldLink parent;
	
	//
	private boolean doInvisible;
	
	
	public FieldLink(ValidatorManager validatorManager,List<Enum<?>> actives,ComboBox box)
	{
		this(validatorManager, actives, box, false);
	}
	
	public FieldLink(ValidatorManager validatorManager,List<Enum<?>> actives,ComboBox box,boolean doInvisible)
	{
		this.validatorManager = validatorManager;
		this.actives = actives;
		this.box = box;
		this.doInvisible = doInvisible;
		
		notNull = new NotNullValidatorConditionnalFieldLink();
		notNull.checkIf(this);
		box.addValueChangeListener(e->valueChanged());
	}
	
	
	public IValidator getValidator()
	{
		return notNull;
	}
	
	public void doLink()
	{
		//
		if (parent!=null)
		{
			throw new AmapjRuntimeException("Cette méthode doit etre appelée uniquement sur la racine");
		}
		//
		initialize();
		
	}
	
	private void initialize()
	{
		currentState = isActif();
		
		for (FieldAndDefaultValue<?> fd : fields)
		{
			AbstractField<?> field = fd.field;
			enableField(field,currentState);

			if (currentState==false)
			{
				fd.setDefaultValueOnDesactivate();
			}
					
			// Si c'est un fieldLink lié, on l'initialize aussi 
			if (fd.fieldLink!=null)
			{
				fd.fieldLink.initialize();
			}
		}
	}
	

	public boolean isActif()
	{
		if (parent!=null && parent.currentState==false)
		{
			return false;
		}
		
		Enum en = (Enum) box.getValue();
		return actives.contains(en);

	}
	
	
	public <T> void  addField(AbstractField<T> f , T defaultValueOnActivate,T defaultValueOnDesactivate)
	{
		addField(f, defaultValueOnActivate, defaultValueOnDesactivate,null);
	}
	
	public <T> void  addField(FieldLink fieldLink , T defaultValueOnActivate,T defaultValueOnDesactivate)
	{
		fieldLink.parent = this;
		addField(fieldLink.box, defaultValueOnActivate, defaultValueOnDesactivate,fieldLink);
	}
	
	
	private <T> void  addField(AbstractField<T> f , T defaultValueOnActivate,T defaultValueOnDesactivate,FieldLink fieldLink)
	{	
		FieldAndDefaultValue<T> fd = new FieldAndDefaultValue<>();
		fd.field = f;
		fd.defaultValueOnActivate = defaultValueOnActivate;
		fd.defaultValueOnDesactivate = defaultValueOnDesactivate;
		fd.fieldLink = fieldLink;

		// Tres important : sinon la desactivation des IntegerTextField ne fonctionne pas
		if (f instanceof TextField && fd.defaultValueOnActivate==null)
		{

			fd.defaultValueOnActivate = (T) "";
		}
		if (f instanceof TextField && fd.defaultValueOnDesactivate==null)
		{

			fd.defaultValueOnDesactivate = (T) "";
		}

		fields.add(fd);
	}
	
	
	public <T> void addField(AbstractField<T> f)
	{
		addField(f,null,null);
	}
	
	

	private void valueChanged()
	{
		boolean previousState = currentState;
		currentState = isActif();
		boolean changeState = previousState!=currentState;
		
		for (FieldAndDefaultValue<?> fd : fields)
		{
			AbstractField<?> field = fd.field;
			
			// Activation - desactivation 
			enableField(field,currentState);

			// Si il y a un  changement d'état ACTIF - INACTIF 
			if (changeState)
			{	
				if (currentState==false)
				{
					fd.setDefaultValueOnDesactivate();
				}
				else
				{
					fd.setDefaultValueOnActivate();
				}
				
				// Si c'est un fieldLink lié, on lui demande de se remetre à jour
				// C'est necessaire car le setDefaultValueOnDesactivate n'est pas suffisant pour activer le valueChanged()
				if (fd.fieldLink!=null)
				{
					fd.fieldLink.valueChanged();
				}
			}
		}
	}

	public List<Enum<?>> getActives()
	{
		return actives;
	}

	public ComboBox getBox()
	{
		return box;
	}

	public String getComboTitle()
	{
		return validatorManager.getTitle(box);
	}
	
	static public class FieldAndDefaultValue<T>
	{
		public AbstractField<T> field;
		
		public T defaultValueOnActivate;
		
		public T defaultValueOnDesactivate;
		
		// Peut être null 
		public FieldLink fieldLink;

		public void setDefaultValueOnDesactivate()
		{
			field.setValue(defaultValueOnDesactivate);
		}
		
		public void setDefaultValueOnActivate()
		{
			field.setValue(defaultValueOnActivate);
		}
	}
	

	
	
	/**
	 * Methode utilitaire pour activer / desactiver un champ
	 */
	private void enableField(AbstractField<?> field,boolean enabled)
	{
		if (doInvisible)
		{
			// Cas des box et des dates qui sont imbriqués dans des HorinontalLayout 
			if ( (field.getParent() instanceof FormLayout)==false)
			{
				field.getParent().setVisible(enabled);
			}
			else
			{
				field.setVisible(enabled);	
			}
		}
		else
		{
			// Activation - desactivation 
			if (field instanceof CKEditorTextField)
			{
				// Attention : il y a un bug dans le wrapper ckeditor, le setEnabled ne fonctionne pas
				( (CKEditorTextField)field).setViewWithoutEditor(!enabled);
			}
			else
			{
				field.setEnabled(enabled);
			}
		}
	}
	
}
